# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from typing import Optional
import os

# Local
from .filelock import FileLock

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------------------ Public methods ------------------------------------------------------------ #

def save(
    path: str,
    content: str,
    encoding: str = 'utf8',
    debug: bool = False
) -> bool:
    try:
        data = None

        if encoding:
            data = content.encode('utf8')
        else:
            data = content

        with open(path, 'wb') as file:
            file.write(data)

            return True
    except Exception as e:
        if debug:
            print(e)

        return False

def save_sync(
    path: str,
    content: str,
    encoding: str = 'utf8',
    timeout: Optional[float] = None,
    debug: bool = False
) -> bool:
    try:
        with FileLock(path, timeout=timeout):
            return save(path=path, content=content, debug=debug)
    except Exception as e:
        if debug:
            print('ERROR - strio.save_sync(\'{}\')'.format(path), e)

        return False

def load(
    path: str,
    encoding: str = 'utf8',
    debug: bool = False
) -> Optional[str]:
    if not os.path.exists(path):
        if debug:
            print('File at \'{}\' does not exist'.format(path))

        return None

    try:
        with open(path, 'rb') as file:
            binary_content = file.read()

        return binary_content if not encoding else binary_content.decode(encoding)
    except Exception as e:
        if debug:
            print(e)

        return None

def load_sync(
    path: str,
    encoding: str = 'utf8',
    timeout: Optional[float] = None,
    debug: bool = False
) -> Optional[str]:
    try:
        with FileLock(path, timeout=timeout):
            return load(path=path, encoding=encoding, debug=debug)
    except Exception as e:
        if debug:
            print('ERROR - strio.load_sync(\'{}\')'.format(path), e)

        return None

# ---------------------------------------------------------------------------------------------------------------------------------------- #


# TAKEN FROM:   https://github.com/ilastik/lazyflow/blob/master/lazyflow/utility/fileLock.py
# AT COMMIT:    https://github.com/ilastik/lazyflow/commit/4c4ec5b63ea668ccdc5f76bbb010c3ccb03fc46f
# COMMITTED ON: Apr 8, 2019

# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from __future__ import print_function
from builtins import range, object

import os, sys, time, errno

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ----------------------------------------------------------- class: FileLock ------------------------------------------------------------ #

class FileLock:
    """ A file locking mechanism that has context-manager support so
        you can use it in a ``with`` statement. This should be relatively cross
        compatible as it doesn't rely on ``msvcrt`` or ``fcntl`` for the locking.
    """

    # ------------------------------------------------------------- Init ------------------------------------------------------------- #    

    class FileLockException(Exception):
        pass

    def __init__(self, protected_file_path, timeout=None, delay=1, lock_file_contents=None):
        """ Prepare the file locker. Specify the file to lock and optionally
            the maximum timeout and the delay between each attempt to lock.
        """
        self.is_locked = False
        self.lockfile = protected_file_path + ".lock"
        self.timeout = timeout
        self.delay = delay
        self._lock_file_contents = lock_file_contents

        if self._lock_file_contents is None:
            self._lock_file_contents = "Owning process args:\n"

            for arg in sys.argv:
                self._lock_file_contents += arg + "\n"


    # -------------------------------------------------------- Public methods -------------------------------------------------------- #

    def locked(self):
        """
        Returns True iff the file is owned by THIS FileLock instance.
        (Even if this returns false, the file could be owned by another FileLock instance, possibly in a different thread or process).
        """
        return self.is_locked

    def available(self):
        """
        Returns True iff the file is currently available to be locked.
        """
        return not os.path.exists(self.lockfile)

    def acquire(self, blocking=True):
        """ Acquire the lock, if possible. If the lock is in use, and `blocking` is False, return False.
            Otherwise, check again every `self.delay` seconds until it either gets the lock or
            exceeds `timeout` number of seconds, in which case it raises an exception.
        """
        start_time = time.time()

        while True:
            try:
                # Attempt to create the lockfile.
                # These flags cause os.open to raise an OSError if the file already exists.
                fd = os.open(self.lockfile, os.O_CREAT | os.O_EXCL | os.O_RDWR)

                with os.fdopen(fd, "a") as f:
                    # Print some info about the current process as debug info for anyone who bothers to look.
                    f.write(self._lock_file_contents)
                break
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise

                if self.timeout is not None and (time.time() - start_time) >= self.timeout:
                    raise FileLock.FileLockException("Timeout occurred.")

                if not blocking:
                    return False

                time.sleep(self.delay)

        self.is_locked = True

        return True

    def release(self):
        """ Get rid of the lock by deleting the lockfile.
            When working in a `with` statement, this gets automatically
            called at the end.
        """
        self.is_locked = False
        os.unlink(self.lockfile)

    def purge(self):
        """
        For debug purposes only.  Removes the lock file from the hard disk.
        """
        if os.path.exists(self.lockfile):
            self.release()

            return True

        return False


    # ------------------------------------------------------- Private methods -------------------------------------------------------- #

    def __enter__(self):
        """ Activated when used in the with statement.
            Should automatically acquire a lock to be used in the with block.
        """
        self.acquire()

        return self

    def __exit__(self, type, value, traceback):
        """ Activated at the end of the with statement.
            It automatically releases the lock if it isn't locked.
        """
        self.release()

    def __del__(self):
        """ Make sure this ``FileLock`` instance doesn't leave a .lock file
            lying around.
        """
        if self.is_locked:
            self.release()


# ---------------------------------------------------------------------------------------------------------------------------------------- #


from typing import Optional, NoReturn
import random, string, time

def string_an(len: int) -> str:
    return ''.join(random.choices(string.ascii_lowercase + string.digits, k=len))

def sleep(
    min_s: float,
    max_s: Optional[float] = None,
    max_deviation_perc: Optional[float] = None
) -> NoReturn:
    multi = 1000.0
    sleep_s = min_s

    if max_s is not None and max_s > 0:
        if max_s < min_s:
            _min_s = min_s
            min_s = max_s
            max_s = min_s
        
        sleep_s = float(random.randint(int(min_s*multi), int(max_s*multi)))/multi
    
    if max_deviation_perc is not None and max_deviation_perc > 0:
        deviation_perc = random.randint(int(-1*max_deviation_perc*multi), int(max_deviation_perc*multi))/multi

        sleep_s += sleep_s*deviation_perc/100
    
    time.sleep(sleep_s)